"""Provides middleware that aims to make request user available globally in the application."""

from django.utils.deprecation import MiddlewareMixin
import threading

local_thread_storage = threading.local()


class DjangoUTZMiddleware(MiddlewareMixin):
    
    def process_request(self, request):
        """Stores the current authenticated user in the local thread storage."""
        if request.user.is_authenticated:
            setattr(local_thread_storage, "django_utz-request_user", request.user)
        else:
            setattr(local_thread_storage, "django_utz-request_user", None)
        

    def process_response(self, request, response):
        """
        Deletes the current authenticated user from the local thread storage.
        """
        if hasattr(local_thread_storage, "django_utz-request_user"):
            delattr(local_thread_storage, "django_utz-request_user")
        return response


def get_request_user():
    """Returns the currently authenticated user. If no user is authenticated, returns None."""
    if hasattr(local_thread_storage, "django_utz-request_user"):
        request_user = getattr(local_thread_storage, "django_utz-request_user")
        if request_user is not None and request_user.is_authenticated:
            return request_user     
    return None


"""django_utz model serializers mixins"""

from typing import Dict
from django.core.exceptions import ImproperlyConfigured

from django_utz.models.mixins import UTZModelMixin
from django_utz.serializers.fields import UTZDateTimeField
from django_utz.utils import is_datetime_field


class UTZModelSerializerMixin:
    """
    ### Automatically adds the fields defined in `self.Meta.model.datetime_fields` to the serializer as `UTZDateTimeField`'s.

    NOTE: This mixin should only be used if the serializer model inherits `UTZModelMixin`. 
    Also, the auto added fields override already present fields with the same name in the serializer.
    To modify the behavior of the auto added fields using keyword arguments, 
    add the keyword arguments in the `extra_kwargs` attribute of the serializer's Meta class.

    If you need complete control of the field's behavior you should not use this mixin and instead add the fields manually to the serializer.

    ```
    from django_utz.serializers.fields import UTZDateTimeField

    class BookSerializer(serializers.ModelSerializer):
        ...
        # Display the `created_at` datetime field as time only in the user's timezone
        created_at_time = UTZDateTimeField(source=created_at, format="%H:%M:%S %Z (%z)", read_only=True)
        updated_at = UTZDateTimeField(format="%Y-%m-%d %H:%M:%S %Z (%z)", read_only=True)
        ...
    ```

    The default format in which the datetime fields are represented in the serializer is `"%Y-%m-%d %H:%M:%S %Z (%z)"`.
    This can be changed by setting the `datetime_repr_format` attribute in the serializer class.

    :attr datetime_repr_format: The format in which the datetime fields should be represented in the serializer.
    Defaults to `"%Y-%m-%d %H:%M:%S %Z (%z)"`.
    :attr add_utz_fields: If True, the fields defined in `self.Meta.model.datetime_fields` will be automatically added to the serializer as `UTZDateTimeField`'s.
    Defaults to True.

    #### Example:
    ```
    from django_utz.serializers.mixins import UTZModelSerializerMixin

    class BookSerializer(UTZModelSerializerMixin, serializers.ModelSerializer):
        datetime_repr_format = "%Y-%m-%d %H:%M:%S %Z"
        
        class Meta:
            model = Book
            fields = "__all__"
            extra_kwargs = {
                "created_at": {"read_only": True},
                "updated_at": {"read_only": True},
            }
    ```
    """
    utz_model_mixin = UTZModelMixin
    add_utz_fields = True
    datetime_repr_format = "%Y-%m-%d %H:%M:%S %Z (%z)"


    def __init__(self, *args, **kwargs):
        obj = super().__init__(*args, **kwargs)
        self._model_serializer_pre_check()
        return obj
    
    
    # Override `get_fields` method to add UTZ fields to the serializer
    def get_fields(self):
        fields = super().get_fields()
        if self.add_utz_fields:
            fields = self.get_and_add_utz_fields(fields)
        return fields


    @property
    def serializer_model(self):
        """
        Returns the serializer's model.

        :return: The serializer's model.
        :rtype: django.db.models.Model
        """
        return self.Meta.model
    

    def _model_serializer_pre_check(self):
        """
        Check if model serializer is properly setup
        """
        assert issubclass(self.utz_model_mixin, UTZModelMixin) == True,(
            f"Invalid `utz_model_mixin` value: {self.utz_model_mixin}."
        )
        if self.utz_model_mixin not in self.serializer_model.__bases__: # Check that the serializer class inherits `self.utz_model_mixin`
            raise ImproperlyConfigured(
                f"Model {self.serializer_model.__name__} does not inherit from {self.utz_model_mixin.__name__}."
            )

        if not hasattr(self.serializer_model, "datetime_fields"):
            raise AttributeError(
                f"Model {self.serializer_model.__class__.__name__} does not have `datetime_fields` attribute."
            )
        return None
    

    def get_utz_field_for_datetime_field(self, datetime_field: str):
        """
        Returns appropriate UTZ Serializer field for the given datetime field name.

        :param datetime_field: The datetime model field for which UTZ serializer field will be returned. Should already be in `self.datetime_fields`
        :return: django_utz.serializers.fields.UTZDateTimeField
        """
        if datetime_field not in self.serializer_model.datetime_fields:
            raise ValueError(f"Invalid property name: {datetime_field}")
        meta_extra_kwargs = self.Meta.extra_kwargs if hasattr(self.Meta, "extra_kwargs") else {}
        extra_kwargs = meta_extra_kwargs.get(datetime_field, {}) # Get extra kwargs for the field if any
        if is_datetime_field(self.serializer_model, datetime_field):
            if "format" not in extra_kwargs:
                extra_kwargs.update({"format": self.datetime_repr_format}) 
            return UTZDateTimeField(**extra_kwargs)
        return None

    
    def get_and_add_utz_fields(self, serializer_fields: Dict):
        """
        Gets and adds to the serializer_fields, a `UTZDateTimeField` for all fields in `self.datetime_fields` 
        that returns fields value adjusted to `self.serializer_model` preferred user's timezone. 
        This overrides any fields already present in the serializer.

        :param serializer_fields: dictionary of fields already present in the serializer.
        :return: serializer_fields with `UTZDateTimeField` for all fields in `self.datetime_fields`.
        """
        assert type(self.serializer_model.datetime_fields) == list, (
            f"Invalid type for `datetime_fields`: {type(self.serializer_model.datetime_fields)}."
        )
        for datetime_field in self.serializer_model.datetime_fields:
            excluded_fields = self.Meta.exclude if hasattr(self.Meta, "exclude") else []
            included_fields = self.Meta.fields if hasattr(self.Meta, "fields") else []
            if (datetime_field in included_fields) and (datetime_field not in excluded_fields):
                serializer_fields[datetime_field] = self.get_utz_field_for_datetime_field(datetime_field)
        return serializer_fields




"""
#### django_utz

This package contains mixins and utils for Django models, DRF model serializers and fields, 
template tags and filters that aid easy conversion of timezone aware fields to the user's timezone.

@Author: Daniel T. Afolayan (ti-oluwa.github.io)
"""

__version__ = "0.1.4"
__author__ = "Daniel T. Afolayan"

alias = "django-user-timezone"

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo
except:
    raise Exception(
        "django_utz couldn't import the zoneinfo module.\
         Perhaps you are on an older version of Python, run `pip install backports.zoneinfo` to continue."
        )



from django.contrib.auth import get_user_model
from django.dispatch import Signal, receiver
from django.db.models.signals import pre_save, post_save

from django_utz.models.mixins import UTZUserModelMixin

User = get_user_model()
user_timezone_changed = Signal(["user", "previous_timezone", "current_timezone"])

@receiver(pre_save, sender=User)
def utz_change_handler_pre_save(sender, **kwargs):
    """High level signal handler for user timezone changes"""
    user = kwargs["instance"]
    if issubclass(user.__class__, UTZUserModelMixin):
        tz_field = getattr(user, "user_timezone_field")
        try:
            previous_timezone = getattr(User.objects.get(pk=user.pk), tz_field)
        except User.DoesNotExist:
            previous_timezone = ''
        current_timezone = getattr(user, tz_field)

        @receiver(post_save, sender=sender)
        def utz_change_handler_post_save(sender, **kwargs):
            """Sends signal when user timezone is changed after it is saved"""
            created = kwargs["created"]
            if not created and previous_timezone != current_timezone:
                user_timezone_changed.send(
                    sender=sender, 
                    user=user, 
                    previous_timezone=previous_timezone, 
                    current_timezone=current_timezone
                )
            return None
    return None


"""Utility functions/classes for django_utz"""
try:
    import zoneinfo
except:
    from backports import zoneinfo
import pytz
import datetime
from django.utils import timezone
from django.db import models
from django.core.exceptions import ValidationError


def validate_timezone(value):
    """
    Validator to check if a timezone is valid.

    :param value: Timezone value to validate
    :raises: ValidationError if the value is not a valid timezone
    :return: The value if it is a valid timezone
    """
    if not is_timezone_valid(value):
        raise ValidationError("Invalid timezone.")
    return value


def is_timezone_valid(timezone: str | datetime.tzinfo):
    """
    Returns whether timezone is a valid timezone.

    :param timezone: timezone info or name
    :return: True if timezone is a valid, False otherwise.
    """
    if isinstance(timezone, datetime.tzinfo):
        return True
    is_pytz = False
    is_zoneinfo = False
    try:
        pytz.timezone(timezone)
        is_pytz = True
    except pytz.exceptions.UnknownTimeZoneError:
        pass
    try:
        zoneinfo.ZoneInfo(timezone)
        is_zoneinfo = True
    except Exception:
        pass
    return is_pytz or is_zoneinfo


def get_attr_by_traversal(obj: object, traversal_path: str, default=None):
    """
    Get an attribute of an object by traversing the object using the traversal path.
    
    :param obj: The object to traverse.
    :param traversal_path: The traversal path to the attribute. For example: "b.c.d"
    :param default: The default value to return if the attribute is not found.

    #### For example:
    ```
    class A:
        b = B()
    
    class B:
        c = C()

    class C:
        d = D()

    class D:
        def __init__(self):
            self.x = 10
    
    traversal_path = "b.c.d.x"
    a = A()
    x = get_attr_by_traversal(a, traversal_path)
    print(x) # 10
    ```
    """
    try:
        attrs = traversal_path.split('.')
        for attr in attrs:
            obj = getattr(obj, attr)
        return obj
    except AttributeError:
        return default


def is_datetime_field(model: models.Model, field_name: str):
    """
    Checks if the given field name is a datetime field in the model.

    :param model: The model to check.
    :param field_name: The name of the field to check.
    :return: True if the field is a datetime field, False otherwise.
    """
    field = model._meta.get_field(field_name)
    return issubclass(field.__class__, models.DateTimeField)


def is_time_field(model: models.Model, field_name: str):
    """
    Checks if the given field name is a time field in the model.

    :param model: The model to check.
    :param field_name: The name of the field to check.
    :return: True if the field is a time field, False otherwise.
    """
    field = model._meta.get_field(field_name)
    return issubclass(field.__class__, models.TimeField)


def is_date_field(model: models.Model, field_name: str):
    """
    Checks if the given field name is a date field in the model.

    :param model: The model to check.
    :param field_name: The name of the field to check.
    :return: True if the field is a date field, False otherwise.
    """
    field = model._meta.get_field(field_name)
    return issubclass(field.__class__, models.DateField)


class utzdatetime(datetime.datetime):
    """Custom datetime class that can be independent of settings.USE_TZ"""
    convert_to_local_time = False

    @classmethod
    def from_datetime(cls, _datetime: datetime.datetime):
        """Returns a utzdatetime object from a datetime.datetime object"""
        if not isinstance(_datetime, datetime.datetime):
            raise TypeError(f"_datetime expected datetime.datetime object, got {type(_datetime)}")
        if timezone.is_naive(_datetime):
            default_timezone = timezone.get_default_timezone()
            _datetime = timezone.make_aware(_datetime, default_timezone)
        return cls(
            year=_datetime.year,
            month=_datetime.month,
            day=_datetime.day,
            hour=_datetime.hour,
            minute=_datetime.minute,
            second=_datetime.second,
            microsecond=_datetime.microsecond,
            tzinfo=_datetime.tzinfo
        )

    def regard_usetz(self):
        """Respect settings.USE_TZ when converting to local time"""
        self.convert_to_local_time = True
        return self

    def disregard_usetz(self):
        """Remain unaffected by settings.USE_TZ when converting to local time"""
        self.convert_to_local_time = False
        return self


"""Provides custom serializer fields for converting datetime to user's local timezone on representation"""
import zoneinfo
from rest_framework import serializers
from django.conf import settings
import pytz
from django.db import models

from django_utz.models.mixins import UTZModelMixin


class UTZBaseField:
    """Base class for UTZ fields"""

    _call_count = 0

    def to_representation(self, value):
        instance = self.root.instance
        # If the instance is a queryset, get the current instance from the queryset
        if issubclass(instance.__class__, (models.QuerySet, list)):
            instance = instance[self._call_count]
            self._call_count += 1
            
        assert issubclass(instance.__class__, UTZModelMixin), (
            f'{self.__class__.__name__} requires the {instance.__class__.__name__} model to inherit from '
            'UTZModelMixin'
        )
        if instance.user_available:
            self.timezone = instance.get_preferred_user().utz
        return super().to_representation(value)
    

    def to_internal_value(self, value):
        if getattr(settings, "USE_DEPRACATED_PYTZ", False):
            server_timezone = pytz.timezone(settings.TIME_ZONE) if settings.USE_TZ else pytz.utc
        else:
            server_timezone = zoneinfo.ZoneInfo(settings.TIME_ZONE) if settings.USE_TZ else zoneinfo.ZoneInfo("UTC")
        return super().to_internal_value(value).astimezone(server_timezone)


class UTZDateTimeField(UTZBaseField, serializers.DateTimeField):
    """
    Custom `serializers.DateTimeField` that converts input datetime to server's timezone(settings.TIMEZONE) before storage
    and converts back to the preferred user's local timezone on representation. 
    
    This field is usually added automatically to the model serializer by the `UTZModelSerializerMixin`. 
    It can also be used as a standalone field in a serializer without the `UTZModelSerializerMixin` mixin.

    if settings.USE_TZ is False, server's timezone is assumed to be UTC

    #### Example:
    ```
    class BookSerializer(serializers.ModelSerializer):
        created_at = UTZDateTimeField(format="%Y-%m-%d %H:%M:%S %Z (%z)")
        updated_at = UTZDateTimeField(format="%Y-%m-%d %H:%M:%S %Z (%z)")

        class Meta:
            model = Book
            fields = "__all__"
            extra_kwargs = {
                "created_at": {"read_only": True},
                "updated_at": {"read_only": True},
            }
    ```
    """
    pass


"""django_utz mixins for models"""

from django.core.exceptions import FieldError, ImproperlyConfigured, FieldDoesNotExist
import types
import pytz
import datetime
from django.db import models
from django.conf import settings
from django.contrib.auth.models import AbstractUser, AbstractBaseUser, User
try:
    import zoneinfo
except:
    from backports import zoneinfo

from django_utz.utils import is_timezone_valid, validate_timezone, get_attr_by_traversal, utzdatetime
from django_utz.middleware import get_request_user


user_models = (AbstractBaseUser, AbstractUser, User)


class UTZUserModelMixin:
    """
    ### This mixin must be used with the user model for the django-utz package to work properly.
    It adds necessary methods and properties to the User model to allow other mixins in this module to work.

    :attr user_timezone_field: The name of the field in the User model which stores the user's timezone. Defaults to None


    #### Below is an example usage of this mixin:
    ```
    from django.db import models
    from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
    from django_utz.models.mixins import UTZUserModelMixin

    class User(UTZUserModelMixin, AbstractBaseUser, PermissionsMixin):
        '''User model'''
        username = models.CharField(max_length=255, unique=True)
        email = models.EmailField(max_length=255, unique=True)
        timezone = models.CharField(max_length=255, blank=True)
        ...

        USERNAME_FIELD = "email"
        REQUIRED_FIELDS = ["username"]

        user_timezone_field = "timezone" # valid user timezone info field name
    ```
    """

    user_timezone_field = None

    def __init__(self, *args, **kwargs):
        obj = super().__init__(*args, **kwargs)
        self._user_model_pre_check()
        return obj


    @property
    def utz(self):
        """
        The user's timezone info as a pytz.tzinfo or zoneinfo.ZoneInfo object.

        if settings.USE_DEPRECATED_PYTZ is True, then the user's timezone info is returned as a pytz.tzinfo object.
        Otherwise, it is returned as a zoneinfo.ZoneInfo object.

        if `user_timezone_field` is not set, then this returns settings.TIME_ZONE or UTC timezone
        """
        if self.user_timezone_field:
            try:
                utz = getattr(self, self.user_timezone_field)
            except AttributeError:
                raise FieldError(
                    f"Field `{self.user_timezone_field}` does not exist in `{self.__class__.__name__}` model.\
                    Please add a `{self.user_timezone_field}` field in the model or set `user_timezone_field` \
                    attribute to the name of the field in `{self.__class__.__name__}` model which stores the user's timezone."
                )
            
            if utz and isinstance(utz, datetime.tzinfo):
                zone_str = utz.__str__()
            elif utz and isinstance(utz, str):
                if not is_timezone_valid(utz):
                    raise ValueError(f"Invalid timezone: {utz}.")
                zone_str = utz
            else:
                zone_str = getattr(settings, "TIME_ZONE", None) or "UTC"

            if getattr(settings, "USE_DEPRECATED_PYTZ", False):
                return pytz.timezone(utz)
        else:
            zone_str = getattr(settings, "TIME_ZONE", None) or "UTC"
        return zoneinfo.ZoneInfo(zone_str)


    @classmethod
    def is_user_model(cls):
        """Check if the model class which contains this mixin is the User model"""
        return issubclass(cls, user_models)
    

    def _user_model_pre_check(self):
        """Ensures that the model in which this mixin is used is the User model"""
        if not self.is_user_model():
            raise TypeError("This mixin can only be used with the User model.")
        
        # Add user_timezone_field to the model if not already added
        if self.user_timezone_field and not hasattr(self, self.user_timezone_field):
            raise ImproperlyConfigured(
                f"Field `{self.user_timezone_field}` does not exist in `{self.__class__.__name__}` model. \
                Please add a `{self.user_timezone_field}` field in the model or set `user_timezone_field` \
                attribute to the name of the field in `{self.__class__.__name__}` model which stores the user's timezone."
            )

        # Add timezone validator to the user_timezone_field if not already added
        try:
            field = self._meta.get_field(self.user_timezone_field)
            if field and validate_timezone not in field.validators:
                field.validators.append(validate_timezone)
        except FieldDoesNotExist:
            pass
        return None 
    

    def to_local_timezone(self, _datetime: datetime.datetime):
        """
        Adjust datetime.datetime object to user's local timezone
        
        :param _datetime: datetime.datetime object
        :return: datetime object adjusted to user's timezone

        datetime object returned is not affected by the USE_TZ setting.
        To change this behavior do:
        ```
        import datetime
        utz_datetime = user.to_local_timezone(datetime.datetime.now())
        utz_datetime.regard_usetz()
        ```
        """
        _utzdatetime = utzdatetime.from_datetime(_datetime)
        user_tz = self.utz
        if user_tz:
            return _utzdatetime.astimezone(user_tz)
        return _datetime



class UTZModelMixin:
    """
    ### Adds properties which return the values of the fields in the `datetime_fields` attribute in the preferred user's timezone.
    #### This properties are created dynamically and added to the model in which this mixin is used. 
    The property names are the field names in the `datetime_fields` attribute plus the suffix specified in the `utz_property_suffix` attribute.

    :attr datetime_fields: The datetime model fields for which the preferred user timezone properties are to be created.
    This fields should be in the model in which this mixin is used. Overriding this attribute is required.
    :attr utz_property_suffix: The suffix to be added to the preferred user timezone properties. Defaults to "utz".
    :attr use_related_user_timezone: If True, the mixin gets the user related to the model(as specified or automatically) and uses its timezone. 
    If False, the mixin uses the request user's timezone. Defaults to False.
    :attr related_user_field_path: The name/traversal path of the field in the model in which this mixin is used, that holds/returns the preferred user.
    :attr _model_superclass: The model superclass. Defaults to `models.Model`.

    #### For Example:
    ```
    # After `UTZUserModelMixin` has been added to the User model
    from django.contrib.auth import get_user_model
    from django_utz.models.mixins import UTZModelMixin

    User = get_user_model()

    class Book(UTZModelMixin, models.Model):
        title = models.CharField(...)
        content = models.TextField(...)
        author = models.ForeignKey(User, ...)
        created_at = models.DateTimeField(...)
        updated_at = models.DateTimeField(...)

        datetime_fields = ["created_at", "updated_at"]
        utz_property_suffix = "local" # defaults to "utz" which stands for "user timezone"
        use_related_user_timezone = False

    # Creating a book object
    book = Book.objects.create(...)
    print(book.created_at_local) # returns the value of `book.created_at` in the request user's timezone

    ```

    If we decide to use the related user timezone instead of the request user's timezone, 
    we can set `use_related_user_timezone` to True and set `related_user_field_path` 
    
    ```
    class Book(UTZModelMixin, models.Model):
        ...
        use_related_user_timezone = True
        related_user_field_path = "author" # The name of the field in the model which returns the preferred user
        
    # Creating a book object
    book = Book.objects.create(...)
    print(book.created_at_local) # returns the value of `book.created_at` in the book.author's timezone FOR ALL USERS
    """
    datetime_fields = []
    utz_property_suffix = "utz"
    use_related_user_timezone = False # If True, the mixin gets the related user object(as specified or automatically) and uses its timezone. If False, the mixin uses the request user's timezone
    related_user_field_path = None
    _model_superclass = models.Model


    def __init__(self, *args, **kwargs):
        obj = super().__init__(*args, **kwargs)
        self._model_pre_check()
        if self.user_available:
            self.create_and_add_utz_properties()
        return obj
    
    @classmethod
    def is_user_model(cls):
        """Check if the model class which contains this mixin is the User model"""
        return issubclass(cls, user_models)

    @property
    def user_available(self):
        """Check if a user object whose timezone is to be used is available"""
        return self.get_preferred_user() is not None
    

    def _model_pre_check(self):
        """
        Check if model is properly setup
        """
        assert issubclass(self._model_superclass, models.Model), (
            f"Invalid superclass for model {self.__class__.__name__}: {self._model_superclass}."
        )
        if not self.is_user_model() and not issubclass(self.__class__, self._model_superclass): # Check that the class inherits `self._model_superclass`
            raise ImproperlyConfigured(
                f"Model {self.__class__.__name__} does not inherit from {self._model_superclass.__name__}."
            )

        if not hasattr(self, "datetime_fields"):
            raise AttributeError(f"Model {self.__class__.__name__} does not have `datetime_fields` attribute.")
        return None
    
    
    def create_and_add_utz_properties(self):
        """
        Creates and adds local timezone properties for all fields in `self.datetime_fields`
        based on the preferred user's timezone.
        """
        assert type(self.datetime_fields) == list, (
            f"Invalid type for `datetime_fields`: {type(self.datetime_fields)}."
        )
        for datetime_field in self.datetime_fields:
            property = types.DynamicClassAttribute(
                fget=self.create_utz_method_for_datetime_field(datetime_field),
                doc=f"Returns the {datetime_field} field's value in the preferred user's timezone."
            )
            setattr(self.__class__, f"{datetime_field}_{self.utz_property_suffix}", property)
        return self


    def get_preferred_user(self):
        """
        Get and returns the user object whose timezone is to be used, based on the `related_user_field_path` attribute if set or
        the request user if `use_related_user_timezone` is False.
        
        If `use_related_user_timezone` is True and `related_user_field_path` is not set, then the mixin
        tries to find the user related field in the model and its related models and returns the first
        user object found.
        """
        user = self
        if not self.is_user_model(): 
            if self.use_related_user_timezone:
                # Get the user object as specified by the `related_user_field_path` attribute or by finding the user related field in the model and its related models
                user_field_traversal_path = self.related_user_field_path or self.find_user_related_model_field()
                if not user_field_traversal_path:
                    if self.related_user_field_path:
                        raise ImproperlyConfigured(
                            f"Please set `related_user_field_path` attribute to the name/traversal \
                            path of the field in {self.__class__.__name__} model which returns the preferred user."
                        )
                    raise FieldError(f"No relation to the User model was found in {self.__class__.__name__}")
                
                user = get_attr_by_traversal(
                    obj=self, 
                    traversal_path=user_field_traversal_path
                ) # Get the user instance using its traversal path from this model
                if not isinstance(user, user_models):
                    raise FieldError(
                        f"object of type {type(user)} from field: '{user_field_traversal_path}'; \
                        is not a User. If you provided a value for `related_user_field_path`, \
                        check that it points to a model field that returns the preferred user"
                    )
            else:
                user = get_request_user()
                
        if self.pk: # Only check if the model object has been saved
            self.check_user_has_utz_mixin(user)
        return user
    

    def create_utz_method_for_datetime_field(self, datetime_field: str):
        """
        Returns a method that would return the datetime model field's value in the user's timezone.

        :param datetime_field: The datetime model field for which the method is to be created.
        should already be in `self.datetime_fields`
        :raises ValueError: If the given datetime_field is not in `self.datetime_fields`
        """
        if datetime_field not in self.datetime_fields:
            raise ValueError(f"Invalid property name: {datetime_field}")
        
        utz_method = f"{datetime_field}_{self.utz_property_suffix}"
        user = self.get_preferred_user()

        def utz_method(self):
            return user.to_local_timezone(getattr(self, datetime_field))
        return utz_method
            
    
    @staticmethod
    def check_user_has_utz_mixin(user: AbstractBaseUser | AbstractUser):
        """
        Checks if the user model inherits the `UTZUserModelMixin` mixin.

        :param user: The user whose model is to be checked.
        """
        if not issubclass(user.__class__, UTZUserModelMixin):
            raise ImproperlyConfigured(f"{user.__class__.__name__} must inherit the `UTZUserModelMixin` mixin.")
        return True
  

    def find_user_related_model_field(self):
        """
        Finds and returns the traversal path to the first user related field found in the model or its related models.

        This method assumes that the user is a foreign key or one-to-one field in the model or its related models.

        :return: The user field traversal path or None if not found.
        """
        field_paths = []

        def find_user_related_field_path(model: models.Model):
            """Finds and returns the user related field traversal path in the given model or its related models."""
            for field in model._meta.fields:
                related_model = field.related_model
                if related_model and isinstance(
                    field, 
                    (models.ForeignKey, models.OneToOneField)
                ):
                    field_paths.append(field.name)
                    if issubclass(related_model, user_models):
                        return ".".join(field_paths)
                    else:
                        # if the field is a relational field but it is related_model is not the user model, 
                        # check the related_model's fields for the user related field
                        return find_user_related_field_path(model=related_model)  
            field_paths.pop()
            return None
                
        return find_user_related_field_path(model=self)

